// ----------------------------------------------------------------------------
//
//  Copyright (C) 2017 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#include "abconfig.h"
#include "oscform.h"


ABconfig::ABconfig (void):
    _dirname (0)
{
    reset ();
}


void ABconfig::reset (void)
{
    _version = 1;
    *_origin = 0;
    *_descript = 0;
    *_micident = 0;
    _lfeq1_f = 100.0f;
    _lfeq1_g = 0.0f;
    _lfeq1_b = 1.0f;
    *_convfile = 0;
}


int ABconfig::save (const char *file)
{
    char          *p;
    FILE          *F;
    char          s [1024];
    Oscform       Z (4096);

    strcpy (s, file);
    p = strrchr (s, '.');
    if (!p) strcat (s, ".octo");
    if (! (F = fopen (s, "w"))) 
    {
	fprintf (stderr, "Can't open '%s' for writing\n", s);
        return 1;
    } 
    
    Z.put_command ("/octoproc");
    Z.put_format (",i");
    Z.put_int (1);
    Z.write (F);

    Z.put_command ("/origin");
    Z.put_format (",s");
    Z.put_string (_origin);
    Z.write (F);

    Z.put_command ("/descript");
    Z.put_format (",s");
    Z.put_string (_descript);
    Z.write (F);

    Z.put_command ("/micident");
    Z.put_format (",s");
    Z.put_string (_micident);
    Z.write (F);

    Z.put_command ("/aformeq/par1");
    Z.put_format (",fff");
    Z.put_float (_lfeq1_f);
    Z.put_float (_lfeq1_g);
    Z.put_float (_lfeq1_b);
    Z.write (F);

    Z.put_command ("/convol/file");
    Z.put_format (",s");
    Z.put_string (_convfile);
    Z.write (F);
	
    fclose (F);
    setdirname (file); // Remember directory name.
    return 0;
}


int ABconfig::load (const char *file)
{
    int           n, err;
    const char    *p, *q;
    FILE          *F;
    char          s [1024];
    Oscform       Z (4096);

    strcpy (s, file);
    p = strrchr (s, '.');
    if (!p) strcat (s, ".octo");
    if (! (F = fopen (s, "r"))) 
    {
	fprintf (stderr, "Can't open file '%s' for reading\n", s);
        return 1;
    } 

    reset (); // Set defaults for any missing data.
    setdirname (file); // Remember directory.

    if (Z.read (F) || ((p = Z.get_command ()) == 0) || strcmp (p, "/octoproc"))
    {
	fprintf (stderr, "File '%s' is not a valid configuration file.\n", s);
        fclose (F);
	return -1;
    }
    if (Z.get_int (&_version) || (_version !=  1))
    {
	fprintf (stderr, "File '%s' has wrong version number.\n", s);
        fclose (F);
	return -1;
    }

    err = 0;
    while (!(err || Z.read (F)))
    {
	p = Z.get_command ();
	if (!p) err = 1;
	else if (!strcmp (p, "/origin"))
	{
	    err = Z.get_string (&q, &n);
	    if (n >= 64) err = -1;
	    if (!err) strcpy (_origin, q);
	}
	else if (!strcmp (p, "/descript"))
	{
	    err = Z.get_string (&q, &n);
	    if (n >= 64) err = -1;
	    if (!err) strcpy (_descript, q);
	}
	else if (!strcmp (p, "/micident"))
	{
	    err = Z.get_string (&q, &n);
	    if (n >= 64) err = -1;
	    if (!err) strcpy (_micident, q);
	}
	else if (!strcmp (p, "/aformeq/par1"))
        {
	    err |= Z.get_float (&_lfeq1_f);
	    err |= Z.get_float (&_lfeq1_g);
	    err |= Z.get_float (&_lfeq1_b);
	}
	else if (!strcmp (p, "/convol/file"))
	{
	    err = Z.get_string (&q, &n);
	    if (n >= 64) err = -1;
	    if (!err) strcpy (_convfile, q);
	}
	else
	{
	    fprintf (stderr, "Command '%s' not understood - ignored'\n",  p);
	}
	if (err)
	{
	    fprintf (stderr, "Format error in command '%s'\n",  p);
	    fclose (F);
	    return -1;
	}
    }

    fclose (F);
    return 0;
}


// Return the full pathname of the convolution file
// for the given sample rate. The returned string,
// if not zero, should be deallocated using free()
// when no longer needed.
//
const char *ABconfig::convolfilepath (int fsamp)
{
    char        *p;
    const char  *s;
    
    if ((_dirname == 0) || (*_convfile == 0)) return 0;
    switch (fsamp)
    {
    case 44100:
        s = "_44.wav";
	break;
    case 48000:
        s = "_48.wav";
	break;
    case 96000:
        s = "_96.wav";
	break;
    default:
	return 0;
    }
    p = (char *) malloc (strlen (_dirname) + strlen (_convfile) + strlen (s) + 1);
    if (p == 0) return 0;
    strcpy (p, _dirname);
    strcat (p, _convfile);
    strcat (p, s);
    return p;
}


int ABconfig::dump (const char *file)
{
    const char *p;

    if (load (file)) return 1;
    printf ("version  : %d\n", _version);
    printf ("origin   : %s\n", _origin);
    printf ("descript : %s\n", _descript);
    printf ("micident : %s\n", _micident);
    printf ("lfeq1    : F = %3.1f Hz,  G = %3.1f dB,  B = %4.2f\n", _lfeq1_f, _lfeq1_g, _lfeq1_b);
    printf ("convfile base name  : %s\n", _convfile);
    p = convolfilepath (44100);
    printf ("convfile (44.1 kHz) : %s\n", p);
    free ((void *) p);
    p = convolfilepath (48000);
    printf ("convfile (48 kHz)   : %s\n", p);
    free ((void *) p);
    p = convolfilepath (96000);
    printf ("convfile (96 kHz)   : %s\n", p);
    free ((void *) p);
    return 0;
}


// The following needs modification for Windows because
// the pathname separator is different, and to allow drive
// names.


#if defined(__linux__) || defined(__FreeBSD_kernel__) || defined(__GNU__)


#include <stdlib.h>
#include <string.h>


// Store the directory name of the config file.
//
void ABconfig::setdirname (const char *file)
{
    char *p;
    
    // Remove any existing directory name.
    if (_dirname)
    {
	free (_dirname);
	_dirname = 0;
    }
    // Handle illegal input.
    if (file == 0) return;
    // Find directory.
    _dirname = strdup (file);
    p = strrchr (_dirname, '/');
    if (p)
    {
	// Keep path up to and including the last '/'.
	p [1] = 0;
    }
    else
    {
	// Assume current directory if no '/' found.
	free (_dirname);
	_dirname = strdup ("./");
    }
}


#endif
