// ----------------------------------------------------------------------------
//
//  Copyright (C) 2017 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <string.h>
#include <assert.h>
#include "convol88.h"


// ----------------------------------------------------------------------------


Convfdata::Convfdata (int npart, int fsize):
    _npart (npart),
    _fsize (fsize)
{
    _freq_data = new fftwf_complex* [npart];
    for (int i = 0; i < npart; i++)
    {
        _freq_data [i] = (fftwf_complex *)(fftwf_malloc (_fsize * sizeof (fftwf_complex)));
    }
}


Convfdata::~Convfdata (void)
{
    for (int i = 0; i < _npart; i++)
    {
        fftwf_free (_freq_data [i]);
    }
    delete[] _freq_data;
}


void Convfdata::clear ()
{
    for (int i = 0; i < _npart; i++)
    {
	memset (_freq_data [i], 0, _fsize * sizeof (fftwf_complex));
    }
}


// ----------------------------------------------------------------------------


Convol88::Convol88 (int maxlen, int period):
    _maxlen (maxlen),
    _period (period),
    _time_data (0),
    _plan_r2c (0),
    _plan_c2r (0)
{
    assert (maxlen > 0);
    assert (((period & (period - 1)) == 0) && (period <= MAXPER));
    if (maxlen > MAXLEN) maxlen  = MAXLEN;

    init ();
    clear ();
    reset ();
}


Convol88::~Convol88 (void)
{
    fini ();
}


void Convol88::init (void)
{
    int i, j;

    _time_data = (float *) (fftwf_malloc (2 * _period * sizeof (float)));
    _freq_data = (fftwf_complex *)(fftwf_malloc ((_period + 1) * sizeof (fftwf_complex)));
    _npart = (_maxlen + _period - 1) / _period;
    _ipart = 0;
    for (i = 0; i < NINP; i++)
    {
	_inp_fdata [i] = new Convfdata (_npart, _period + 1);
        for (j = 0; j < NOUT; j++)
	{
	    _mac_fdata [i][j] = new Convfdata (_npart, _period + 1);
	}
    }
    for (j = 0; j < NOUT; j++)
    {
	_outbuff [j] = new float [_period];
    }
    _plan_r2c = fftwf_plan_dft_r2c_1d (2 * _period, _time_data, _freq_data, 0);
    _plan_c2r = fftwf_plan_dft_c2r_1d (2 * _period, _freq_data, _time_data, 0);
}


void Convol88::fini (void)
{
    int i, j;

    fftwf_free (_time_data);
    fftwf_free (_freq_data);
    for (i = 0; i < NINP; i++)
    {
	delete _inp_fdata [i];
        for (j = 0; j < NOUT; j++)
	{
	    delete _mac_fdata [i][j];
	}
    }
    for (j = 0; j < NOUT; j++)
    {
	delete[] _outbuff [j];
    }
    fftwf_destroy_plan (_plan_r2c);
    fftwf_destroy_plan (_plan_c2r);
}


int Convol88::setimp (int inp, int out, float gain, float *data, int size, int step)
{
    int     i, j, n;
    Convfdata *fdata;

    if ((inp < 0) || (inp >= NINP)) return 1;
    if ((out < 0) || (out >= NOUT)) return 1;
    if (size > _maxlen) size = _maxlen;

    gain /= 2 * _period;
    fdata = _mac_fdata [inp][out];
    fdata->clear ();
    for (i = 0; size; i++)
    {
	memset (_time_data, 0, 2 * _period * sizeof (float));
	n = size;
	if (n > _period) n = _period;
	for (j = 0; j < n; j++)
	{
	    _time_data [j] = data [j * step] * gain;
	}
        fftwf_execute_dft_r2c (_plan_r2c, _time_data, fdata->_freq_data [i]);	
	data += n * step;
	size -= n;
    }
    return 0;
}


void Convol88::reset (void)
{
    int i, j;

    for (i = 0; i < NINP; i++) _inp_fdata [i]->clear ();
    for (j = 0; j < NOUT; j++) memset (_outbuff [j], 0, _period * sizeof (float));
}


void Convol88::clear (void)
{
    int i, j;

    for (i = 0; i < NINP; i++)
    {
	for (j = 0; j < NOUT; j++)
	{
	    _mac_fdata [i][j]->clear ();
	}
    }
}


void Convol88::process (float *inp [], float *out [])
{
    int            i, j, k, ia, ib;
    float          *p, *q;;
    fftwf_complex  *A, *B;


    for (i = 0; i < NINP; i++)
    {
	memcpy (_time_data, inp [i], _period * sizeof (float));
	memset (_time_data + _period, 0, _period * sizeof (float));
	fftwf_execute_dft_r2c (_plan_r2c, _time_data, _inp_fdata [i]->_freq_data [_ipart]);	
    }

    for (j = 0; j < NOUT; j++)
    {
	memset (_freq_data, 0, (_period + 1) * sizeof (fftwf_complex));
	for (i = 0; i < NINP; i++)
	{
	    ia = _ipart;
	    for (ib = 0; ib < _npart; ib++)
	    {
		A = _inp_fdata [i]->_freq_data [ia];
		B = _mac_fdata [i][j]->_freq_data [ib];
		for (k = 0; k <= _period; k++)
		{
		    _freq_data [k][0] += A [k][0] * B [k][0] - A [k][1] * B [k][1];
		    _freq_data [k][1] += A [k][0] * B [k][1] + A [k][1] * B [k][0];
		}
		if (--ia < 0) ia += _npart;
	    }
	}
	fftwf_execute_dft_c2r (_plan_c2r, _freq_data, _time_data);	
	p = out [j];
	q = _outbuff [j];
	for (k = 0; k < _period; k++) p [k] = q [k] + _time_data [k];
	memcpy (q, _time_data + _period, _period * sizeof (float));
    }

    if (++_ipart == _npart) _ipart = 0;
}


// ----------------------------------------------------------------------------
