// ----------------------------------------------------------------------------
//
//  Copyright (C) 2017-2018 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <stdio.h>
#include <math.h>
#include "audiofile.h"
#include "abconfig.h"
#include "abprocess.h"
#include "global.h"


enum
{
    HELP,
    HPF, INV, SN3D, N3D, FUMA,
    CAF, WAV, AMB, AIFF, BIT16, BIT24, FLOAT, REC, TRI, LIPS 
};

enum { BUFFSIZE = 1024 };


static float hpfreq = 40.0f;
static int   shnorm = SH_SN3D;
static bool  invmic = false;
static int   type = Audiofile::TYPE_CAF;
static int   form = Audiofile::FORM_24BIT;
static int   dith = Audiofile::DITHER_NONE;


static void help (void)
{
    fprintf (stderr, "\noctofile %s\n", VERSION);
    fprintf (stderr, "(C) 2017-2018 Fons Adriaensen  <fons@linuxaudio.org>\n");
    fprintf (stderr, "A to B format conversion for the Core Sound Octomic.\n");
    fprintf (stderr, "Usage: octofile <options> <calibration file> <input file(s)> <output file>.\n");
    fprintf (stderr, "Options:\n");
    fprintf (stderr, "  Display this text:     --help\n");
    fprintf (stderr, "  Highpass filter:       --hpf <frequency>\n");
    fprintf (stderr, "  Upside down mic:       --inv\n");
    fprintf (stderr, "  B format standard:     --sn3d, --n3d, --fuma\n");
    fprintf (stderr, "  Output file type:      --caf, --wav, --amb, --aiff\n");
    fprintf (stderr, "  Output sample format:  --16bit, --24bit, --float\n");
    fprintf (stderr, "  Dither type (16 bit):  --rec, --tri, --lips\n");
    fprintf (stderr, "There can be 1, 2, 4 or 8 input file(s).\n");
    fprintf (stderr, "The default options are --hpf 40 --sn3d --caf --24bit.\n");
    exit (1);
}


static struct option options [] = 
{
    { "help",  0, 0, HELP  },
    { "hpf",   1, 0, HPF   },
    { "inv",   0, 0, INV   },
    { "sn3d",  0, 0, SN3D  },
    { "n3d",   0, 0, N3D   },
    { "fuma",  0, 0, FUMA  },
    { "caf",   0, 0, CAF   },
    { "wav",   0, 0, WAV   },
    { "amb",   0, 0, AMB   },
    { "aiff",  0, 0, AIFF  },
    { "16bit", 0, 0, BIT16 },
    { "24bit", 0, 0, BIT24 },
    { "float", 0, 0, FLOAT },
    { "rec",   0, 0, REC   },
    { "tri",   0, 0, TRI   },
    { "lips",  0, 0, LIPS  },
    { 0, 0, 0, 0 }
};


static void procoptions (int ac, char *av [])
{
    int k;

    while ((k = getopt_long (ac, av, "", options, 0)) != -1)
    {
        switch (k)
        {
        case '?':
        case HELP:
            help ();
            break;
        case HPF:
            if ((sscanf (optarg, "%f", &hpfreq) != 1) || (hpfreq < 10) || (hpfreq > 320))
            {
                fprintf (stderr, "Illegal value for --hpf option: '%s'.\n", optarg);
                exit (1);
            }
            break;
        case INV:
            invmic = true;
            break;
        case SN3D:
            shnorm = SH_SN3D;
            break;
        case N3D:
            shnorm = SH_N3D;
            break;
        case FUMA:
            shnorm = SH_FUMA;
            break;
        case CAF:
            type = Audiofile::TYPE_CAF;
            break;
        case WAV:
            type = Audiofile::TYPE_WAV;
            break;
        case AMB:
            type = Audiofile::TYPE_AMB;
	    shnorm = SH_FUMA;
            break;
        case AIFF:
            type = Audiofile::TYPE_AIFF;
            break;
        case BIT16:
            form = Audiofile::FORM_16BIT;
            break;
        case BIT24:
            form = Audiofile::FORM_24BIT;
            break;
        case FLOAT:
            form = Audiofile::FORM_FLOAT;
            break;
        case REC:
            dith = Audiofile::DITHER_RECT;
            break;
        case TRI:
            dith = Audiofile::DITHER_TRIA;
            break;
        case LIPS:
            dith = Audiofile::DITHER_LIPS;
            break;
        }
    }
}


int main (int ac, char *av [])
{
    Audiofile     Ainp [8];
    Audiofile     Aout;
    ABprocess     abproc;
    ABconfig      config;
    int           i, j, k, n, ninp, rate = 0, rv = 1;
    float         *buff, *inp [8], *out [9];
    float         *p;
    
    // Process command line options.
    procoptions (ac, av);

    // Check number of input files.
    if (optind == ac - 1) return config.dump (av [optind]);
    ninp = ac - optind - 2;
    if (ninp < 1)
    {
        fprintf (stderr, "Missing arguments, try --help.\n");
        return 1;
    }
    if ((ninp != 1) && (ninp != 2) && (ninp != 4) && (ninp != 8))
    {
        fprintf (stderr, "There must be 1, 2, 4 or 8 input files.\n");
        return 1;
    }

    // Check conflicting options.
    if ((type == Audiofile::TYPE_AMB) && (shnorm != SH_FUMA))
    {
        fprintf (stderr, "File type 'amb' requires FuMa format.\n");
        return 1;
    }
		
    // Load configuration file.
    // Error message is printed by config.load().
    if (config.load (av [optind])) return 1;
    optind++;
    
    // Open and check input files.
    for (j = 0; j < ninp; j++)
    {    
        if (Ainp [j].open_read (av [optind]))
        {
            fprintf (stderr, "Can't open input file '%s'.\n", av [optind]);
            return 1;
        }
        if (Ainp [j].chan () != 8 / ninp)
        {
            fprintf (stderr, "Input file '%s' should have %d channel(s).\n", av [optind], 8 / ninp);
            goto cleanup;
        }
        optind++;
        if (j == 0)
        {
            rate = Ainp [0].rate ();
            continue;
        }
        if (Ainp [j].rate () != rate)
        {
            fprintf (stderr, "Sample rates of input files don't match.\n");
            goto cleanup;
        }
        if (Ainp [j].size () != Ainp [0].size ())
        {
            fprintf (stderr, "Sizes of input files don't match.\n");
            goto cleanup;
        }
    }

    // Open output file.
    if (Aout.open_write (av [optind], type, form, rate, 9))
    {
        fprintf (stderr, "Can't open output file '%s'.\n", av [optind]);
        goto cleanup;
    }
    if (dith != Audiofile::DITHER_NONE) 
    {
        Aout.set_dither (dith);
    }

    // Initialise the A/B processor.
    abproc.init (rate, BUFFSIZE);

    // Apply the configuration.
    if (abproc.load_config (&config))
    {
        fprintf (stderr, "Configuration failed\n");
        goto cleanup;
    }

    // Set user options.
    abproc.set_hipfil (hpfreq);
    abproc.set_invmic (invmic);
    abproc.set_shnorm (shnorm);
    
    // Allocate processing buffers.
    // Interleaved file data.
    buff = new float [BUFFSIZE * 9];
    // Single channel in and out.
    for (i = 0; i < 8; i++)
    {
        inp [i] = new float [BUFFSIZE];
    }
    for (i = 0; i < 9; i++)
    {
        out [i] = new float [BUFFSIZE];
    }

    // Process until end of file.
    while (true)
    {
        // Read input files, de-interleave if necessary.
        if (ninp == 8)
        {
            for (j = 0; j < 8; j++)
            {
                k = Ainp [j].read (inp [j], BUFFSIZE);
            }
        }
        else if (ninp == 4)
        {
            for (j = n = 0; j < 4; j++, n += 2)
            {
                k = Ainp [j].read (buff, BUFFSIZE);
                for (i = 0, p = buff; i < k; i++, p += 2)
                {
                    inp [n + 0][i] = p [0];
                    inp [n + 1][i] = p [1];
                }
            }
        }
        else if (ninp == 2)
        {
            for (j = n = 0; j < 2; j++, n += 4)
            {
                k = Ainp [j].read (buff, BUFFSIZE);
                for (i = 0, p = buff; i < k; i++, p += 4)
                {
                    inp [n + 0][i] = p [0];
                    inp [n + 1][i] = p [1];
                    inp [n + 2][i] = p [2];
                    inp [n + 3][i] = p [3];
                }
            }
        }
        else
        {
            k = Ainp [0].read (buff, BUFFSIZE);
            for (i = 0, p = buff; i < k; i++, p += 8)
            {
                inp [0][i] = p [0];
                inp [1][i] = p [1];
                inp [2][i] = p [2];
                inp [3][i] = p [3];
                inp [4][i] = p [4];
                inp [5][i] = p [5];
                inp [6][i] = p [6];
                inp [7][i] = p [7];
            }
        }

        if (k > 0)
        {
            if (k < BUFFSIZE)
            {
                // Add zeros if not full buffer.
                n = BUFFSIZE - k;
                for (j = 0; j < 8; j++)
                {
                    memset (inp [j] + k, 0, n * sizeof (float));
                }
            }
            // Process buffer.
            abproc.process (inp, out); 
            // Interleave and write to output file.
            for (i = 0, p = buff; i < k; i++, p += 9)
            {
                if (shnorm == SH_FUMA)
                {
		    // Permute channels for FuMa format.
                    p [0] = out [0][i];
                    p [1] = out [3][i];
                    p [2] = out [1][i];
                    p [3] = out [2][i];
                    p [4] = out [6][i];
                    p [5] = out [7][i];
                    p [6] = out [5][i];
                    p [7] = out [8][i];
                    p [8] = out [4][i];
                }
                else
                {
		    // Use ACN channel order.
                    p [0] = out [0][i];
                    p [1] = out [1][i];
                    p [2] = out [2][i];
                    p [3] = out [3][i];
                    p [4] = out [4][i];
                    p [5] = out [5][i];
                    p [6] = out [6][i];
                    p [7] = out [7][i];
                    p [8] = out [8][i];
                }
            }
            Aout.write (buff, k);
        }
        else
        {
            // End of input.
            break;
        }
    }

    // Done, close output file and dealloc buffers.
    Aout.close ();
    delete[] buff;
    for (j = 0; j < 8; j++) delete[] inp [j];
    for (j = 0; j < 9; j++) delete[] out [j];
    rv = 0;

    // Cleanup input files.
cleanup:
    for (j = 0; j < ninp; j++) Ainp [j].close ();

    return rv;
}
